C
C
C **********************************************************************
C
C     Linear Optimal Control Problem with Bang-bang Structure
C     due to Plant and Athans (1966)
C
C     D.H. Jacobson, D.Q. Mayne: Differential Dynamic Programming;
C     Elsevier, 1970 (pp. 85-88).
C
C     S.J. Wright: Interior Point Methods for Optimal Control
C     of Discrete Time Systems; J.Optim. Theory a. Appl.,
C     Vol. 77, pp.161-187, 1993 (Example 6.3).
C
C     R. Pytlak: Numerical Methods for Optimal Control Problems
C     with State Constraints; Springer, 1999 (pp. 72-74).    
C
C     Minimize J = x1(tf)^2 + .. + x4(tf)^2, tf = 4.2,  subject to
C
C     x1' = -0.5*x1 + 5*x2,       x1(0) = 10,
C     x2' = -5*x1 - 0.5*x2 + u,   x2(0) = 10,
C     x3' = -0.6*x3 + 10*x4,      x3(0) = 10,
C     x4' = -10*x3 - 0.6*x4 + u,  x4(0) = 10,
C
C     -1 <= u <= 1.
C
C     Due to Jacobson, Mayne the solution is of bang-bang type with
C     seven switching points
C
C     In a first step we freeze the control (with estimated switching
C     points) solving a BVP for the state and adjoint variables;
C     in a second step we set the switching points free demanding
C     the switching function to vanish at these points.
C
C **********************************************************************
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      PARAMETER (MMAX=11,MSMAX=8,MMS=19,NMS=16,
     *           NDW=NMS*(8+MMAX*(10+6*NMS)),NDIW=5*NMS)
      DIMENSION X(MMS),XS(MSMAX),Y(NMS,MMS),WORK(NDW)
      INTEGER JS(MMS,MSMAX),IWORK(NDIW)
      EXTERNAL F,R,DIFSYB,RKF7B
      COMMON /ITEIL/ ITEIL
      COMMON /PARA/  U1, SCHALT, HAM
C
      OPEN(6,FILE='athans.txt',STATUS='UNKNOWN')
      OPEN(7,FILE='athans.dat',STATUS='UNKNOWN')
      OPEN(10,FILE='athans11.dat',STATUS='UNKNOWN')
      OPEN(11,FILE='athans12.dat',STATUS='UNKNOWN')
C
C --- Parameters
C
      N      =  8
      M      =  11
      MS     =  7
      TOL    =  1.D-09
      NFILE  =  6
      KS     =  1
C
C --- Initial Trajektory
C
      DO 100 K=1,M
      X(K)   = 4.2D0*DFLOAT(K-1)/DFLOAT(M-1)
      DO 110 I=1,4
      Y(I,K) = 10.D0
  110 CONTINUE
      DO 120 I=5,8
      Y(I,K) = 1.D0
  120 CONTINUE
  100 CONTINUE
C
      XS(1) = 0.1D0
      XS(2) = 0.9D0
      XS(3) = 1.4D0
      XS(4) = 2.2D0
      XS(5) = 2.6D0
      XS(6) = 3.4D0
      XS(7) = 3.9D0      
C
C --- Solution of the Boundary Value Problem 
C     with freezed control function 
C
      ITEIL  =  1
      KP     =  0
      ITMAX  =  20
      CALL BNDSCO(F,R,DIFSYB,X,XS,Y,WORK,IWORK,JS,N,M,MS,
     *     KS,TOL,ITMAX,KP,MMAX,MSMAX,MMS,NMS,NDW,NDIW,NFILE)
      IF(KP.LT.0) GOTO 900
C
C --- Solution with free switching points 
C
      ITEIL  =  2
      CALL RESET(X,XS,Y,N,M,MS,NMS)
      KP     =  0
      ITMAX  =  20
      CALL BNDSCO(F,R,DIFSYB,X,XS,Y,WORK,IWORK,JS,N,M,MS,
     *     KS,TOL,ITMAX,KP,MMAX,MSMAX,MMS,NMS,NDW,NDIW,NFILE)
      IF(KP.LT.0) GOTO 900
C
C --- Performance Index
C
      XJ     =  Y(1,M)**2 + Y(2,M)**2 + Y(3,M)**2 + Y(4,M)**2
      WRITE(6,2001) XJ
C 
C
C --- Numerical Integration of the Solution Trajectory,
C     Determination of the Data for Plot with MATLAB
C     Files: athans.dat, athans.m
C 
      ITEIL  =  3
      KP     =  0
      IFEIN  =  5
      IPLOT  =  1
      NPLOT  =  7
      ICASE  =  1
      CALL AWP(DIFSYB,F,X,XS,Y,N,M,MS,
     *         JS,MMS,NMS,IFEIN,IPLOT,KP,NFILE,NPLOT,ICASE)
C
C     Storage of the Solution Data 
C     (The subroutine RESET removes the switching points)
C
      CALL RESET(X,XS,Y,N,M,MS,NMS)
      DO 810 K=1,M
      WRITE(11,1001) X(K)
      WRITE(11,1001) (Y(I,K),I=1,N)
  810 CONTINUE
      WRITE(11,1001) (XS(K),K=1,MS)
C
  900 CONTINUE
C
 1001 FORMAT(4D20.12)
 2001 FORMAT(//' Performance Index:',D22.12)
C -------------------------------------------------------------------
      STOP
      END




      SUBROUTINE F(X,Y,DY,J,L,JS,MMS)
C
C     Linear Optimal Control Problem with Bang-bang Structure
C     due to Plant and Athans (1966)
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      DIMENSION Y(*),DY(*)
      INTEGER JS(MMS,*)
      COMMON /ITEIL/ ITEIL
      COMMON /PARA/  U1, SCHALT, HAM
C
C     Optimal Control
C
      U1     =  -1.D0
      IF(JS(J,1)*JS(J,2).LT.0) U1 = 1.D0
      IF(JS(J,3)*JS(J,4).LT.0) U1 = 1.D0
      IF(JS(J,5)*JS(J,6).LT.0) U1 = 1.D0
      IF(JS(J,7).GT.0) U1 = 1.D0
C
      SCHALT = Y(6) + Y(8)
C 
C     State - and Adjoint Equations
C
      DY(1)  =  -0.5D0*Y(1) + 5.D0*Y(2)
      DY(2)  =  -5.D0*Y(1) - 0.5D0*Y(2) + U1
      DY(3)  =  -0.6D0*Y(3) + 10.D0*Y(4)
      DY(4)  =  -10.D0*Y(3) - 0.6D0*Y(4) + U1
C
      DY(5)  =   0.5D0*Y(5) + 5.D0*Y(6)
      DY(6)  =  -5.D0*Y(5) + 0.5D0*Y(6)
      DY(7)  =   0.6D0*Y(7) + 10.D0*Y(8)
      DY(8)  =  -10.D0*Y(7) + 0.6D0*Y(8)
C
C     Hamiltonian
C     
      HAM    =  Y(5)*DY(1)+Y(6)*DY(2)+Y(7)*DY(3)+Y(8)*DY(4)
C -------------------------------------------------------------------
      RETURN
      END




      SUBROUTINE R(YA,YB,ZZ,W,NYA,NSK,J,L,LS,JS,MMS)
C
C     Linear Optimal Control Problem with Bang-bang Structure
C     due to Plant and Athans (1966)
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      DIMENSION YA(*),YB(*),ZZ(*),W(*),DV(7)
      INTEGER NYA(*),NSK(*)
      INTEGER JS(MMS,*)
      COMMON /ITEIL/ ITEIL
      COMMON /PARA/  U1, SCHALT, HAM
C
      W(1)   =  YA(1)/10.D0 - 1.D0
      W(2)   =  YA(2)/10.D0 - 1.D0
      W(3)   =  YA(3)/10.D0 - 1.D0
      W(4)   =  YA(4)/10.D0 - 1.D0
      W(5)   =  2.D0*YB(1) - YB(5)
      W(6)   =  2.D0*YB(2) - YB(6)
      W(7)   =  2.D0*YB(3) - YB(7)
      W(8)   =  2.D0*YB(4) - YB(8)
C
      NYA(1) =  1
      NYA(2) =  2
      NYA(3) =  3
      NYA(4) =  4
C
C     Freezing the Switching Points
C 
      IF(ITEIL.EQ.1) THEN  
      W(9)   =  0.D0
      W(10)  =  0.D0
      W(11)  =  0.D0
      W(12)  =  0.D0
      W(13)  =  0.D0
      W(14)  =  0.D0
      W(15)  =  0.D0
C
      NYA(9) =  9
      NYA(10)=  10
      NYA(11)=  11
      NYA(12)=  12
      NYA(13)=  13
      NYA(14)=  14
      NYA(15)=  15
      ENDIF
C
C     Switching Conditions for free Switching Points
C
      IF(ITEIL.EQ.2) THEN  
      W(9)   =  SCHALT
      W(10)  =  SCHALT
      W(11)  =  SCHALT
      W(12)  =  SCHALT
      W(13)  =  SCHALT
      W(14)  =  SCHALT
      W(15)  =  SCHALT
C
      NSK(9) =  1
      NSK(10)=  2
      NSK(11)=  3
      NSK(12)=  4
      NSK(13)=  5
      NSK(14)=  6
      NSK(15)=  7
      ENDIF
C  
C -------------------------------------------------------------------
      RETURN
      END



      SUBROUTINE AWP(METHOD,FCN,X,XS,Y,N,M,MS,JS,MMS,NMS,IFEIN,
     *               IPLOT,KP,NFILE,NPLOT,ICASE)
C
C *******************************************************************
C
C     INTEGRATION OF INITIAL VALUE PROBLEMS
C
C     IFEIN : NUMBER OF INTEGRATION NODES IN SUBINTERVAL
C     IPLOT : THE NODES ARE STORED FOR PLOT SUBROUTINE,
C             IF IPLOT.EQ.1 THE ROUTINE TAKES IFEIN = 40
C             EQUIDISTANT POINTS IN EACH SUBINTERVAL
C     KP    : PRINT PARAMETER
C             KP.LT.0 : NO PRINT
C             KP.EQ.0 : MULTIPLE SHOOTING NODES
C             KP.GT.0 : ALL INTEGRATION NODES
C     ICASE : 1  :  ROUTINE INTEGRATES OUTPUT OF BNDSCO IN EACH
C                   SUBINTERVAL (THE SWITCHING POINTS ARE ALREADY
C                   ARRANGED IN ORDER; THE ROUTINE TAKES NEW INITIAL
C                   DATA FOR EACH SUBINTERVAL)
C             0  :  INTEGRATION OVER THE WHOLE INTERVAL; THE ALGORITHM
C                   ARRANGES THE SWITCHING POINTS XS(K) IN ORDER AND
C                   SAVES THE DATA Y(I;J);
C
C *******************************************************************
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      DIMENSION X(*),XS(*),Y(NMS,*)
      DIMENSION Z(20),DZ(20)
      INTEGER JS(MMS,*)
      EXTERNAL FCN , R
      COMMON /ITEIL/ ITEIL
      COMMON /PARA/  U1, SCHALT, HAM
C
      IF(IPLOT.GT.0) IFEIN=40
      TOL=1.D-07
C
      NN  =  N
C
C     ARRANGING THE SWITCHING POINTS (IN CASE ICASE = 0)
C
      IF(MS.EQ.0.OR.ICASE.GT.0) GOTO 300
      K   =  1
  200 CONTINUE
      K1  =  K + 1
      X1  =  X(K)
      X2  =  X(K1)
      DO 210 L=1,MS
      IF((X1-XS(L))*(X2-XS(L)).LT.-1.D-10) GOTO 230
  210 CONTINUE
      GOTO 250
  230 CONTINUE
      M   =  M + 1
      DO 240 J=M,K+2,-1
      X(J) = X(J-1)
  240 CONTINUE
      X(K1) = XS(L)
      GOTO 200
  250 K   =  K1
      IF(K.LT.M) GOTO 200
  300 CONTINUE
C
      JT  =  0
      L   =  0
      HI  =  0.D0
      M1  =  M - 1
      IF(KP.GE.0) THEN
        WRITE(NFILE,2000)
        WRITE(NFILE,2200)
      ENDIF
C
C     LOOP OVER THE SUBINTERVALS
C
      DO 500 K=1,M1
      J=K
      J1=J+1
      L=L+1
C
      T=X(J)
      TEND=X(J1)
      IF(MS.GT.0) THEN
         DO 503 KL=1,MS
  503    IF(T.EQ.XS(KL)) L=L-1
      ENDIF
      HMAX=(TEND-T)/IFEIN
      IF(HI.EQ.0.D0) HI=0.01D0*HMAX
      HMAX=ABS(HMAX)
C
C     DEFINITION OF JS(J,K), K=1,..,MS (ICASE = 0)
C
      IF(ICASE.EQ.0) THEN
      DO 410 KL=1,MS
      JS(J,KL) = 1
      IF(T.LT.XS(KL)-1.D-10) JS(J,KL) = -1
  410 CONTINUE
      END IF
C
C     NEW INITIAL DATA FOR THE INTEGRATION
C
      IF(ICASE.EQ.0.AND.J.GT.1) GOTO 430
      DO 420 I=1,N
  420 Z(I)=Y(I,J)
  430 CONTINUE
C
      CALL FCN(T,Z,DZ,J,L,JS,MMS)
C
      IF(KP.GE.0) WRITE(NFILE,2001)
      JT=JT+1
      IF(KP.GE.0)
     *WRITE(NFILE,3000) T,(Z(I),I=1,NN),SCHALT,U1,HAM
      IF(IPLOT.EQ.1)
     *WRITE(NPLOT,4000) T,(Z(I),I=1,NN),SCHALT,U1,HAM
C
C     LOOP FOR SUB-SUB-INTERVALS
C
      DO 510 K1=1,IFEIN
      T1=T+SIGN(HMAX,HI)
      IF(K1.EQ.IFEIN) T1=TEND
C
      CALL METHOD(NN,FCN,T,Z,T1,TOL,HMAX,HI,J,L,JS,MMS)
C
      IF(HI.EQ.0.D0) GOTO 600
C
      CALL FCN(T,Z,DZ,J,L,JS,MMS)
C
      JT=JT+1
      IF(KP.GT.0)
     *WRITE(NFILE,3000) T,(Z(I),I=1,NN),SCHALT,U1,HAM
      IF(IPLOT.EQ.1)
     *WRITE(NPLOT,4000) T,(Z(I),I=1,NN),SCHALT,U1,HAM
C
  510 CONTINUE
C
C     DEFINITION OF THE MULTIPLE SHOOTING KNOTS (ICASE = 0)
C
      IF(ICASE.EQ.0) THEN
      DO 520 I=1,NN
      Y(I,J+1) = Z(I)
  520 CONTINUE
      ENDIF
C
  500 CONTINUE
C
      IF(KP.EQ.0) THEN
      WRITE(NFILE,2001)
      WRITE(NFILE,3000) T,(Z(I),I=1,NN),SCHALT,U1,HAM
      ENDIF
      IF(KP.GE.0) WRITE(NFILE,2001)
      GOTO 900
C
  600 CONTINUE
      WRITE(NFILE,5000)
      WRITE(NFILE,3000) T,(Z(I),I=1,NN),SCHALT,U1,HAM
C
  900 CONTINUE
 1000 FORMAT(I4)
 2000 FORMAT(///)
 2001 FORMAT(' ',77('*'))
 2200 FORMAT(7X,'t',17X,'x1',17X,'x2',17X,'x3'/
     *       25X,'x4',17X,'l1',17X,'l2'/
     *       25X,'l3',17X,'l4',17X,'S'/
     *       25X,'U1',17X,'Ham')
 3000 FORMAT(4D19.12/5(19X,3D19.12/))
 4000 FORMAT(5(30E13.5/))
 5000 FORMAT(' INTEGRATIONSABBRUCH: SCHRITTWEITE H = 0')
C -------------------------------------------------------------------
      RETURN
      END



      SUBROUTINE RESET(X,XS,Y,N,M,MS,NMS)
C
      IMPLICIT DOUBLE PRECISION (A-H,O-Z)
      DIMENSION X(*),XS(*),Y(NMS,*)
C
      J1=0
      MD=M
      DO 600 J2=1,M
      DO 601 I=1,MS
      IF(XS(I).EQ.X(J2)) GOTO 620
  601 CONTINUE
      J1=J1+1
      X(J1)=X(J2)
      DO 610 I=1,N
  610 Y(I,J1)=Y(I,J2)
      GOTO 600
  620 MD=MD-1
  600 CONTINUE
      M=MD
C -------------------------------------------------------------------
      RETURN
      END

